@@ -1,24 +1,3 @@
-from operator import attrgetter
-from functools import wraps
-import socket
-import time
-import sys
-import re
-import os
-
-from collections.abc import MutableSequence
-
-from ciscoconfparse.protocol_values import ASA_TCP_PORTS, ASA_UDP_PORTS
-import ciscoconfparse
-from dns.exception import DNSException
-from dns.resolver import Resolver
-from dns import reversename, query, zone
-
-from ipaddress import IPv4Network, IPv6Network, IPv4Address, IPv6Address
-import ipaddress
-
-from loguru import logger
-
 r""" ccp_util.py - Parse, Query, Build, and Modify IOS-style configurations
 
      Copyright (C) 2021-2022 David Michael Pennington
@@ -43,6 +22,31 @@ r""" ccp_util.py - Parse, Query, Build, and Modify IOS-style configurations
      mike [~at~] pennington [/dot\] net
 """
 
+from operator import attrgetter
+from functools import wraps
+import socket
+import time
+import sys
+import re
+import os
+
+from collections.abc import MutableSequence
+from ipaddress import IPv4Network, IPv6Network, IPv4Address, IPv6Address
+from ipaddress import collapse_addresses as ipaddr_collapse_addresses
+from ipaddress import AddressValueError
+
+
+from dns.exception import DNSException
+from dns.resolver import Resolver
+from dns import reversename, query, zone
+
+from loguru import logger
+
+from ciscoconfparse.protocol_values import ASA_TCP_PORTS, ASA_UDP_PORTS
+from ciscoconfparse.errors import PythonOptimizeException
+import ciscoconfparse
+
+
 # Maximum ipv4 as an integer
 IPV4_MAXINT = 4294967295
 # Maximum ipv6 as an integer
@@ -115,8 +119,71 @@ class UnsupportedFeatureWarning(SyntaxWarning):
     pass
 
 
+class PythonOptimizeCheck(object):
+    """
+    Check if we're running under "python -O ...".  The -O option removes
+    all `assert` statements at runtime.  ciscoconfparse depends heavily on
+    `assert` and running ciscoconfparse under python -O is a really bad idea.
+
+    __debug__ is True unless run with `python -O ...`.  __debug__ is False
+    under `python -O ...`.
+
+    Also throw an error if PYTHONOPTIMIZE is set in the windows or unix shell.
+
+    This class should be run in <module_name_dir>/__init__.py.
+
+    This condition is not unique to ciscoconfparse.
+
+    Simple usage (in __init__.py):
+    ------------------------------
+
+    # Handle PYTHONOPTIMIZE problems...
+    from ciscoconfparse.ccp_util import PythonOptimizeCheck
+    _ = PythonOptimizeCheck()
+
+
+    """
+    @logger.catch(reraise=True)
+    def __init__(self):
+
+        self.PYTHONOPTIMIZE_env_value = os.environ.get("PYTHONOPTIMIZE", None)
+
+        error = "__no_error__"
+        try:
+            # PYTHONOPTIMIZE is not supported...  in the linux shell
+            # disable it with `unset PYTHONOPTIMIZE`
+            if isinstance(self.PYTHONOPTIMIZE_env_value, str) and self.PYTHONOPTIMIZE_env_value.strip()!="":
+                # This condition explicitly allows PYTHONOPTIMIZE="", which
+                # is not a problem.
+                error = "Your environment has PYTHONOPTIMIZE set.  ciscoconfparse doesn't support running under PYTHONOPTIMIZE."
+            # PYTHONOPTIMIZE is not supported...  in the linux shell
+            # disable it with `unset PYTHONOPTIMIZE`
+            elif self.PYTHONOPTIMIZE_env_value is not None:
+                error = "Your environment has PYTHONOPTIMIZE set.  ciscoconfparse doesn't support running under PYTHONOPTIMIZE."
+            # Throw an error if we're running under `python -O`.  `python -O` is not supported
+            # We should keep the __debug__ check for `-O` at the end, otherwise it
+            # masks identifying problems with PYTHONOPTIMIZE set in the shell...
+            elif __debug__ is False:
+                # Running under 'python -O'
+                error = "You're using `python -O`. Please don't.  ciscoconfparse doesn't support `python -O`"
+
+            else:
+                # whew...
+                pass
+
+        except Exception as exception_info:
+            print("exception_info", str(exception_info))
+            raise RuntimeError("Something bad happened in PYTHONOPTIMIZE checks.  Please report this problem as a ciscoconfparse bug")
+
+        if error != "__no_error__":
+            raise PythonOptimizeException(error)
+
+
+@logger.catch(reraise=True)
 def as_text_list(object_list):
-    """This is a helper-function to convert a list of configuration objects into a list of text config lines.
+    """
+    This is a helper-function to convert a list of configuration objects into
+    a list of text config lines.
 
     Examples
     --------
@@ -142,7 +209,7 @@ def as_text_list(object_list):
     >>>
 
     """
-    assert isinstance(object_list, list) or isinstance(object_list, tuple)
+    assert isinstance(object_list, (list, tuple,))
     for obj in object_list:
         assert isinstance(obj.linenum, int)
         assert isinstance(obj.text, str)
@@ -150,9 +217,11 @@ def as_text_list(object_list):
     return list(map(attrgetter("text"), object_list))
 
 
+@logger.catch(reraise=True)
 def junos_unsupported(func):
     """A function wrapper to warn junos users of unsupported features"""
 
+    @logger.catch(reraise=True)
     def wrapper(*args, **kwargs):
         warn = "syntax='junos' does not fully support config modifications such as .{}(); see Github Issue #185.  https://github.com/mpenning/ciscoconfparse/issues/185".format(
             func.__name__
@@ -171,6 +240,7 @@ def junos_unsupported(func):
     return wrapper
 
 
+@logger.catch(reraise=True)
 def log_function_call(function=None, *args, **kwargs):
     """A wrapper; this decorator uses loguru to log function calls.
 
@@ -183,6 +253,7 @@ def log_function_call(function=None, *args, **kwargs):
 
     """
 
+    @logger.catch(reraise=True)
     def logging_decorator(ff):
         @wraps(ff)
         def wrapped_logging(*args, **kwargs):
@@ -196,10 +267,11 @@ def log_function_call(function=None, *args, **kwargs):
                         "Type 2 log_function_call: %s(%s, %s)"
                         % (ff.__qualname__, args, kwargs)
                     )
-            return ff(*args, **kwargs)
+
             logger.info(
                 f"Type 3 log_function_call: {ff.__qualname__}({args}, {kwargs})"
             )
+            return ff(*args, **kwargs)
 
         return wrapped_logging
 
@@ -211,44 +283,44 @@ def log_function_call(function=None, *args, **kwargs):
     return logging_decorator
 
 
+@logger.catch(reraise=True)
 def ccp_logger_control(
     sink=sys.stderr,
     action="",
     handler_id=None,
     allow_enqueue=True,
-    rotation="00:00",
-    retention="1 month",
-    compression="zip",
+    # rotation="00:00",
+    # retention="1 month",
+    # compression="zip",
     level="DEBUG",
+    colorize=True,
     debug=0,
 ):
-    """A simple function to handle logging... Enable / Disable all ciscoconfparse logging here... also see Github issue #211.
+    """
+    A simple function to handle logging... Enable / Disable all
+    ciscoconfparse logging here... also see Github issue #211.
 
     Example
     -------
     """
 
-    msg = "ccp_logger_control() was called with sink='{}', action='{}', handler_id='{}', allow_enqueue={}, rotation='{}', retention='{}', compression='{}', level='{}', debug={}".format(
+    msg = "ccp_logger_control() was called with sink='{}', action='{}', handler_id='{}', allow_enqueue={}, level='{}', colorize={}, debug={}".format(
         sink,
         action,
         handler_id,
         allow_enqueue,
-        rotation,
-        retention,
-        compression,
+        # rotation,
+        # retention,
+        # compression,
         level,
+        colorize,
         debug,
     )
-    logger.info(msg)
+    if debug > 0:
+        logger.info(msg)
 
     assert isinstance(action, str)
-    assert (
-        action == "remove"
-        or action == "add"
-        or action == "disable"
-        or action == "enable"
-        or action == ""
-    )
+    assert action in ("remove", "add", "disable", "enable", "",)
 
     package_name = "ciscoconfparse"
 
@@ -273,19 +345,18 @@ def ccp_logger_control(
 
         logger.add(
             sink=sink,
-            colorize=True,
             diagnose=True,
             backtrace=True,
             # https://github.com/mpenning/ciscoconfparse/issues/215
             enqueue=allow_enqueue,
             serialize=False,
             catch=True,
-            rotation="00:00",
-            retention="1 day",
-            compression="zip",
+            # rotation="00:00",
+            # retention="1 day",
+            # compression="zip",
+            colorize=True,
             level="DEBUG",
         )
-        # format='<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | <level>{level: <8}</level> | <cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> - <level>{message}</level>'
         logger.enable(package_name)
         return True
 
@@ -300,9 +371,11 @@ def ccp_logger_control(
         )
 
 
-@logger.catch(default=True, onerror=lambda _: sys.exit(1))
-class __ccp_re__:
-    """A wrapper around python's re.  This is an experimental object... it may disappear at any time as long as this message exists.
+class __ccp_re__(object):
+    """
+    A wrapper around python's re.  This is an experimental object... it may
+    disappear at any time as long as this message exists.
+
     self.regex = r'{}'.format(regex)
     self.compiled = re.compile(self.regex, flags=flags)
     self.group = group
@@ -329,7 +402,9 @@ class __ccp_re__:
 
     """
 
-    def __init__(self, regex_str=r"", target_str=None, groups=None, flags=0, debug=0):
+    @logger.catch(reraise=True)
+    def __init__(self, regex_str=r"", target_str=None, groups=None, flags=0,
+        debug=0):
         assert isinstance(regex_str, str)
         assert isinstance(flags, int)
         assert isinstance(debug, int)
@@ -349,16 +424,20 @@ class __ccp_re__:
         self.groups = groups
         self.search_result = None
 
+    # do NOT wrap with @logger.catch(...)
     def __repr__(self):
         return f"""ccp_re({self.regex}, {self.target_str})"""
 
+    # do NOT wrap with @logger.catch(...)
     def __str__(self):
         return f"""ccp_re({self.regex}, {self.target_str})"""
 
+    # do NOT wrap with @logger.catch(...)
     @property
     def regex(self):
         return r"""%s""" % self.regex_str
 
+    # do NOT wrap with @logger.catch(...)
     @regex.setter
     def regex(self, regex_str):
         assert isinstance(regex_str, str)
@@ -366,6 +445,8 @@ class __ccp_re__:
         self.compiled = re.compile(regex_str)
         self.attempted_search = False
 
+
+    # do NOT wrap with @logger.catch(...)
     def s(self, target_str):
         assert self.attempted_search is False
         assert isinstance(target_str, str)
@@ -382,10 +463,12 @@ class __ccp_re__:
         else:
             return None
 
+    # do NOT wrap with @logger.catch(...)
     @property
     def result(self):
         raise NotImplementedError()
 
+    # do NOT wrap with @logger.catch(...)
     @property
     def captured(self):
         rv_groups = list()
@@ -432,6 +515,7 @@ class __ccp_re__:
         return rv_groups, rv_groupdict
 
 
+# do NOT wrap with @logger.catch(...)
 def _get_ipv4(val="", strict=False, stdlib=False, debug=0):
     """Return the requested IPv4 object to the caller.  This method heavily depends on IPv4Obj()"""
     assert isinstance(val, str) or isinstance(val, int)
@@ -457,9 +541,10 @@ def _get_ipv4(val="", strict=False, stdlib=False, debug=0):
                 assert isinstance(obj.network, IPv4Network)
                 return obj.network
     except Exception as ee:
-        raise ipaddress.AddressValueError(str(ee))
+        raise AddressValueError("_get_ipv4(val='%s')" % (val))
 
 
+# do NOT wrap with @logger.catch(...)
 def _get_ipv6(val="", strict=False, stdlib=False, debug=0):
     """Return the requested IPv6 object to the caller.  This method heavily depends on IPv6Obj()"""
     assert isinstance(val, str) or isinstance(val, int)
@@ -486,8 +571,9 @@ def _get_ipv6(val="", strict=False, stdlib=False, debug=0):
                 return obj.network
 
     except Exception as ee:
-        raise ipaddress.AddressValueError(str(ee))
+        raise AddressValueError("_get_ipv6(val='%s')" % (val))
 
+# do NOT wrap with @logger.catch(...)
 def ip_factory(val="", stdlib=False, mode="auto_detect", debug=0):
     """
     Accept an IPv4 or IPv6 address / (mask or masklength).  Return an appropriate IPv4 or IPv6 object
@@ -502,7 +588,6 @@ def ip_factory(val="", stdlib=False, mode="auto_detect", debug=0):
     assert isinstance(stdlib, bool)
     assert isinstance(debug, int)
 
-
     obj = None
     if mode == "auto_detect":
 
@@ -521,7 +606,7 @@ def ip_factory(val="", stdlib=False, mode="auto_detect", debug=0):
             return obj
         else:
             error_str = "Cannot auto-detect ip='%s'" % val
-            raise ipaddress.AddressValueError(error_str)
+            raise AddressValueError(error_str)
 
     elif mode == "ipv4":
         try:
@@ -529,7 +614,7 @@ def ip_factory(val="", stdlib=False, mode="auto_detect", debug=0):
             return obj
         except:
             error_str = "Cannot parse '%s' as ipv4" % val
-            raise ipaddress.AddressValueError(error_str)
+            raise AddressValueError(error_str)
 
     elif mode == "ipv6":
         try:
@@ -537,21 +622,13 @@ def ip_factory(val="", stdlib=False, mode="auto_detect", debug=0):
             return obj
         except:
             error_str = "Cannot parse '%s' as ipv6" % val
-            raise ipaddress.AddressValueError(error_str)
+            raise AddressValueError(error_str)
 
     else:
         error_str = "Cannot parse '%s' as ipv4 or ipv6" % val
-        raise ipaddress.AddressValueError(error_str)
-
-    # Raise errors for any problems...  We should not be here...
-    error = "ip_factory('{}', mode='{}') could not parse into a valid {} object".format(
-        val,
-        mode,
-        addr_family,
-    )
-    raise ipaddress.AddressValueError(error)
-
+        raise AddressValueError(error_str)
 
+@logger.catch(reraise=True)
 def collapse_addresses(network_list):
     """
     This is a ciscoconfparse proxy for ipaddress.collapse_addresses()
@@ -565,6 +642,7 @@ def collapse_addresses(network_list):
     """
     assert isinstance(network_list, list) or isinstance(network_list, tuple)
 
+    @logger.catch(reraise=True)
     def ip_net(arg):
         if isinstance(arg, IPv4Obj):
             return arg.network
@@ -575,15 +653,18 @@ def collapse_addresses(network_list):
         elif isinstance(arg, IPv6Network):
             return arg
         else:
-            ValueError("collapse_addresses() isn't sure how to handle %s" % arg)
+            raise ValueError("collapse_addresses() isn't sure how to handle %s" % arg)
 
-    return ipaddress.collapse_addresses([ip_net(ii) for ii in network_list])
+    return ipaddr_collapse_addresses([ip_net(ii) for ii in network_list])
 
 
 # Build a wrapper around ipaddress classes to mimic the behavior of network
 # interfaces (such as persisting host-bits when the intf masklen changes) and
 # add custom @properties
-class IPv4Obj:
+class IPv4Obj(object):
+
+    # This method is on IPv4Obj().  Do NOT add @logger.catch to __init__()...
+    # that breaks it.
     def __init__(self, arg=f"127.0.0.1/{IPV4_MAX_PREFIXLEN}", strict=False, debug=0):
         """An object to represent IPv4 addresses and IPv4 networks.
 
@@ -712,7 +793,11 @@ class IPv4Obj:
         self.params_dict = {}
 
         # Build params_dict... this needs to work with any supported input...
-        if isinstance(arg, str) or isinstance(arg, int):
+        if isinstance(arg, str):
+            params_dict = self._ipv4_params_dict(arg)
+            self.params_dict = params_dict
+
+        elif isinstance(arg, int):
             params_dict = self._ipv4_params_dict(arg)
             self.params_dict = params_dict
 
@@ -761,30 +846,26 @@ class IPv4Obj:
             return None
 
         else:
-            raise ipaddress.AddressValueError(
+            raise AddressValueError(
                 "Could not parse '{}' (type: {}) into an IPv4 Address".format(
                     arg, type(arg)
                 )
             )
 
+    # do NOT wrap with @logger.catch(...)
     # On IPv4Obj()
     def _ipv4_params_dict(self, arg, debug=0):
-        """Parse out important IPv4 parameters from arg.  This method must run to completion for address parsing to work correctly."""
+        """
+        Parse out important IPv4 parameters from arg.  This method must run to
+        completion for IPv4 address parsing to work correctly.
+        """
         assert isinstance(arg, str) or isinstance(arg, int) or isinstance(arg, IPv4Obj)
 
-        params_dict = {
-            'ipv4_addr': None,
-            'ip_version': 4,
-            'ip_arg_str': None,
-            'netmask': None,
-            'masklen': None,
-        }
-
         if isinstance(arg, str):
             try:
                 mm = _RGX_IPV4ADDR_NETMASK.search(arg)
             except TypeError:
-                raise ipaddress.AddressValueError(
+                raise AddressValueError(
                     f"_ipv4_params_dict() doesn't understand how to parse {arg}"
                 )
 
@@ -811,7 +892,7 @@ class IPv4Obj:
             elif masklen is not None:
                 ip_arg_str = f"{addr}/{masklen}"
             else:
-                raise ipaddress.AddressValueError()
+                raise AddressValueError()
 
         elif isinstance(arg, int):
             addr = str(IPv4Address(arg))
@@ -826,7 +907,7 @@ class IPv4Obj:
             ip_arg_str = f"{addr}/{masklen}"
 
         else:
-            raise ipaddress.AddressValueError("IPv4Obj(arg='%s') is an unknown argument type" % (arg))
+            raise AddressValueError("IPv4Obj(arg='%s') is an unknown argument type" % (arg))
 
         assert 0 <= masklen <= IPV4_MAX_PREFIXLEN
         params_dict = {
@@ -839,11 +920,13 @@ class IPv4Obj:
 
         return params_dict
 
+    # do NOT wrap with @logger.catch(...)
     # On IPv4Obj()
     def __repr__(self):
         assert isinstance(self.prefixlen, int)
         return f"""<IPv4Obj {str(self.ip_object)}/{self.prefixlen}>"""
 
+    # do NOT wrap with @logger.catch(...)
     # On IPv4Obj()
     def __eq__(self, val):
         try:
@@ -865,10 +948,12 @@ class IPv4Obj:
             )
             raise AttributeError(errmsg)
 
+    # do NOT wrap with @logger.catch(...)
     # On IPv4Obj()
     def __ne__(self, val):
         return not self.__eq__(val)
 
+    # do NOT wrap with @logger.catch(...)
     # On IPv4Obj()
     def __gt__(self, val):
         try:
@@ -903,6 +988,7 @@ class IPv4Obj:
             errmsg = f"{self.__repr__()} cannot compare itself to '{val}'"
             raise ValueError(errmsg)
 
+    # do NOT wrap with @logger.catch(...)
     # On IPv4Obj()
     def __lt__(self, val):
         try:
@@ -938,6 +1024,7 @@ class IPv4Obj:
             logger.error(errmsg)
             raise ValueError(errmsg)
 
+    # do NOT wrap with @logger.catch(...)
     # On IPv4Obj()
     def __int__(self):
         """Return this object as an integer"""
@@ -946,6 +1033,7 @@ class IPv4Obj:
         else:
             return False
 
+    # do NOT wrap with @logger.catch(...)
     # On IPv4Obj()
     def __index__(self):
         """Return this object as an integer (used for hex() and bin() operations)"""
@@ -954,6 +1042,7 @@ class IPv4Obj:
         else:
             return False
 
+    # do NOT wrap with @logger.catch(...)
     # On IPv4Obj()
     def __add__(self, val):
         """Add an integer to IPv4Obj() and return an IPv4Obj()"""
@@ -968,6 +1057,7 @@ class IPv4Obj:
         retval.prefixlen = orig_prefixlen
         return retval
 
+    # do NOT wrap with @logger.catch(...)
     # On IPv4Obj()
     def __sub__(self, val):
         """Subtract an integer from IPv4Obj() and return an IPv4Obj()"""
@@ -982,6 +1072,7 @@ class IPv4Obj:
         retval.prefixlen = orig_prefixlen
         return retval
 
+    # do NOT wrap with @logger.catch(...)
     # On IPv4Obj()
     def __contains__(self, val):
         # Used for "foo in bar"... python calls bar.__contains__(foo)
@@ -1011,25 +1102,30 @@ class IPv4Obj:
                 )
             )
 
+    # do NOT wrap with @logger.catch(...)
     # On IPv4Obj()
     def __hash__(self):
         # Python3 needs __hash__()
         return hash(str(self.ip_object)) + hash(str(self.prefixlen))
 
+    # do NOT wrap with @logger.catch(...)
     # On IPv4Obj()
     def __iter__(self):
         return self.network_object.__iter__()
 
+    # do NOT wrap with @logger.catch(...)
     # On IPv4Obj()
     def __next__(self):
         ## For Python3 iteration...
         return self.network_object.__next__()
 
+    # do NOT wrap with @logger.catch(...)
     # On IPv4Obj()
     def next(self):
         ## For Python2 iteration...
         return self.network_object.__next__()
 
+    # do NOT wrap with @logger.catch(...)
     # On IPv4Obj()
     @property
     def _version(self):
@@ -1038,6 +1134,7 @@ class IPv4Obj:
         """
         return self.version
 
+    # do NOT wrap with @logger.catch(...)
     # On IPv4Obj()
     @property
     def _prefixlen(self):
@@ -1046,6 +1143,7 @@ class IPv4Obj:
         """
         return self.prefixlen
 
+    # do NOT wrap with @logger.catch(...)
     # On IPv4Obj()
     @property
     def _max_prefixlen(self):
@@ -1057,35 +1155,41 @@ class IPv4Obj:
         else:
             return IPV6_MAX_PREFIXLEN
 
+    # do NOT wrap with @logger.catch(...)
     # On IPv4Obj()
     @staticmethod
     def get_regex():
         return _IPV4_REGEX_STR
 
+    # do NOT wrap with @logger.catch(...)
     # On IPv4Obj()
     @property
     def _ip(self):
         """Returns the address as an integer.  This property exists for compatibility with ipaddress.IPv4Address() in stdlib"""
         return int(self.ip_object)
 
+    # do NOT wrap with @logger.catch(...)
     # On IPv4Obj()
     @property
     def ip(self):
         """Returns the address as an :class:`ipaddress.IPv4Address` object."""
         return self.ip_object
 
+    # do NOT wrap with @logger.catch(...)
     # On IPv4Obj()
     @property
     def netmask(self):
         """Returns the network mask as an :class:`ipaddress.IPv4Address` object."""
         return self.network_object.netmask
 
+    # do NOT wrap with @logger.catch(...)
     # On IPv4Obj()
     @property
     def masklen(self):
         """Returns the length of the network mask as an integer."""
         return int(self.network_object.prefixlen)
 
+    # do NOT wrap with @logger.catch(...)
     # On IPv4Obj()
     @masklen.setter
     def masklen(self, arg):
@@ -1094,17 +1198,20 @@ class IPv4Obj:
             f"{str(self.ip_object)}/{arg}", strict=False
         )
 
+    # do NOT wrap with @logger.catch(...)
     # On IPv4Obj()
     @property
     def netmask(self):
         return self.network_object.netmask
 
+    # do NOT wrap with @logger.catch(...)
     # On IPv4Obj()
     @property
     def masklength(self):
         """Returns the length of the network mask as an integer."""
         return self.prefixlen
 
+    # do NOT wrap with @logger.catch(...)
     # On IPv4Obj()
     @masklength.setter
     def masklength(self, arg):
@@ -1113,12 +1220,14 @@ class IPv4Obj:
             f"{str(self.ip_object)}/{arg}", strict=False
         )
 
+    # do NOT wrap with @logger.catch(...)
     # On IPv4Obj()
     @property
     def prefixlen(self):
         """Returns the length of the network mask as an integer."""
         return int(self.network_object.prefixlen)
 
+    # do NOT wrap with @logger.catch(...)
     # On IPv4Obj()
     @prefixlen.setter
     def prefixlen(self, arg):
@@ -1127,12 +1236,14 @@ class IPv4Obj:
             f"{str(self.ip_object)}/{arg}", strict=False
         )
 
+    # do NOT wrap with @logger.catch(...)
     # On IPv4Obj()
     @property
     def prefixlength(self):
         """Returns the length of the network mask as an integer."""
         return self.prefixlen
 
+    # do NOT wrap with @logger.catch(...)
     # On IPv4Obj()
     @prefixlength.setter
     def prefixlength(self, arg):
@@ -1141,18 +1252,21 @@ class IPv4Obj:
             f"{str(self.ip_object)}/{arg}", strict=False
         )
 
+    # do NOT wrap with @logger.catch(...)
     # On IPv4Obj()
     @property
     def exploded(self):
         """Returns the IPv4 Address object as a string.  The string representation is in dotted decimal notation. Leading zeroes are never included in the representation."""
         return self.ip_object.exploded
 
+    # do NOT wrap with @logger.catch(...)
     # On IPv4Obj()
     @property
     def packed(self):
         """Returns the IPv4 object as packed hex bytes"""
         return self.ip_object.packed
 
+    # do NOT wrap with @logger.catch(...)
     # On IPv4Obj()
     @property
     def broadcast(self):
@@ -1162,6 +1276,7 @@ class IPv4Obj:
         else:
             return self.network_object.broadcast_address
 
+    # do NOT wrap with @logger.catch(...)
     # On IPv4Obj()
     @property
     def network(self):
@@ -1181,30 +1296,35 @@ class IPv4Obj:
     #        [int(num, 16) * (65536 ** idx) for idx, num in enumerate(num_strings)]
     #    )
 
+    # do NOT wrap with @logger.catch(...)
     # On IPv4Obj()
     @property
     def hostmask(self):
         """Returns the host mask as an :class:`ipaddress.IPv4Address` object."""
         return self.network_object.hostmask
 
+    # do NOT wrap with @logger.catch(...)
     # On IPv4Obj()
     @property
     def max_int(self):
         """Return the maximum size of an IPv4 Address object as an integer"""
         return IPV4_MAXINT
 
+    # do NOT wrap with @logger.catch(...)
     # On IPv4Obj()
     @property
     def inverse_netmask(self):
         """Returns the host mask as an :class:`ipaddress.IPv4Address` object."""
         return self.network_object.hostmask
 
+    # do NOT wrap with @logger.catch(...)
     # On IPv4Obj()
     @property
     def version(self):
         """Returns the IP version of the object as an integer.  i.e. 4"""
         return 4
 
+    # do NOT wrap with @logger.catch(...)
     # On IPv4Obj()
     @property
     def numhosts(self):
@@ -1214,6 +1334,7 @@ class IPv4Obj:
         else:
             return 2 ** (IPV4_MAX_PREFIXLEN - self.network_object.prefixlen)
 
+    # do NOT wrap with @logger.catch(...)
     # On IPv4Obj()
     @property
     def as_decimal(self):
@@ -1222,6 +1343,7 @@ class IPv4Obj:
         num_strings.reverse()  # reverse the order
         return sum(int(num) * (256**idx) for idx, num in enumerate(num_strings))
 
+    # do NOT wrap with @logger.catch(...)
     # On IPv4Obj()
     @property
     def as_decimal_network(self):
@@ -1230,12 +1352,14 @@ class IPv4Obj:
         num_strings.reverse()  # reverse the order
         return sum(int(num) * (256**idx) for idx, num in enumerate(num_strings))
 
+    # do NOT wrap with @logger.catch(...)
     # On IPv4Obj()
     @property
     def as_int(self):
         """Returns the IP address as a decimal integer"""
         return self.as_decimal
 
+    # do NOT wrap with @logger.catch(...)
     # On IPv4Obj()
     @property
     def as_zeropadded(self):
@@ -1243,6 +1367,7 @@ class IPv4Obj:
         num_strings = str(self.ip).split(".")
         return ".".join([f"{int(num):03}" for num in num_strings])
 
+    # do NOT wrap with @logger.catch(...)
     # On IPv4Obj()
     @property
     def as_zeropadded_network(self):
@@ -1254,30 +1379,35 @@ class IPv4Obj:
             + str(self.prefixlen)
         )
 
+    # do NOT wrap with @logger.catch(...)
     # On IPv4Obj()
     @property
     def as_hex(self):
         """Returns the IP address as a hex string"""
         return hex(self)
 
+    # do NOT wrap with @logger.catch(...)
     # On IPv4Obj()
     @property
     def as_binary_tuple(self):
         """Returns the IP address as a tuple of zero-padded binary strings"""
         return tuple(f"{int(num):08b}" for num in str(self.ip).split("."))
 
+    # do NOT wrap with @logger.catch(...)
     # On IPv4Obj()
     @property
     def as_hex_tuple(self):
         """Returns the IP address as a tuple of zero-padded hex strings"""
         return tuple(f"{int(num):02x}" for num in str(self.ip).split("."))
 
+    # do NOT wrap with @logger.catch(...)
     # On IPv4Obj()
     @property
     def as_cidr_addr(self):
         """Returns a string with the address in CIDR notation"""
         return str(self.ip) + "/" + str(self.prefixlen)
 
+    # do NOT wrap with @logger.catch(...)
     # On IPv4Obj()
     @property
     def as_cidr_net(self):
@@ -1287,18 +1417,21 @@ class IPv4Obj:
         else:
             return str(self.network)
 
+    # do NOT wrap with @logger.catch(...)
     # On IPv4Obj()
     @property
     def is_multicast(self):
         """Returns a boolean for whether this is a multicast address"""
         return self.network_object.is_multicast
 
+    # do NOT wrap with @logger.catch(...)
     # On IPv4Obj()
     @property
     def is_private(self):
         """Returns a boolean for whether this is a private address"""
         return self.network_object.is_private
 
+    # do NOT wrap with @logger.catch(...)
     # On IPv4Obj()
     @property
     def is_reserved(self):
@@ -1309,7 +1442,10 @@ class IPv4Obj:
 # Build a wrapper around ipaddress classes to mimic the behavior of network
 # interfaces (such as persisting host-bits when the intf masklen changes) and
 # add custom @properties
-class IPv6Obj:
+class IPv6Obj(object):
+
+    # This method is on IPv6Obj().  Do NOT add @logger.catch to __init__()...
+    # that breaks it.
     def __init__(self, arg=f"::1/{IPV6_MAX_PREFIXLEN}", strict=False, debug=0):
         """An object to represent IPv6 addresses and IPv6 networks.
 
@@ -1413,31 +1549,22 @@ class IPv6Obj:
             return None
 
         else:
-            raise ipaddress.AddressValueError(
-                "Could not parse '{}' (type: {}) into an IPv6 Address".format(
-                    arg, type(arg)
-                )
-            )
+            raise AddressValueError("IPv6Obj(arg='%s') is an unknown argument type" % (arg))
 
     # On IPv6Obj()
     def _ipv6_params_dict(self, arg, debug=0):
-        """Parse out important IPv6 parameters from arg.  This method must run to completion for address parsing to work correctly."""
+        """
+        Parse out important IPv6 parameters from arg.  This method must run to
+        completion for IPv6 address parsing to work correctly.
+        """
         assert isinstance(arg, str) or isinstance(arg, int) or isinstance(arg, IPv6Obj)
 
-        params_dict = {
-            'ipv6_addr': None,
-            'ip_version': 6,
-            'ip_arg_str': None,
-            'netmask': None,
-            'masklen': None,
-        }
-
         if isinstance(arg, str):
             try:
                 mm = _RGX_IPV6ADDR.search(arg)
 
             except TypeError:
-                raise ipaddress.AddressValueError(
+                raise AddressValueError(
                     f"_ipv6_params_dict() doesn't know how to parse {arg}"
                 )
 
@@ -1465,32 +1592,30 @@ class IPv6Obj:
             addr = str(IPv6Address(arg))
             netmask = 'ffff:ffff:ffff:ffff:ffff:ffff:ffff:ffff'
             masklen = 128
-            ip_arg_str = f"{addr}/{masklen}"
 
         elif isinstance(arg, IPv6Obj):
             addr = str(arg.ip)
             netmask = str(arg.netmask)
             masklen = arg.masklen
-            ip_arg_str = f"{addr}/{masklen}"
 
         else:
-            raise ipaddress.AddressValueError("IPv6Obj(arg='%s') is an unknown argument type" % (arg))
+            raise AddressValueError("IPv6Obj(arg='%s')" % (arg))
 
         assert 0 <= masklen <= IPV6_MAX_PREFIXLEN
+
         params_dict = {
             'ipv6_addr': addr,
             'ip_version': 6,
-            'ip_arg_str': None,
+            'ip_arg_str': str(addr) + "/" + str(masklen),
             'netmask': netmask,
             'masklen': masklen,
         }
 
         if params_dict.get('masklen', None) is not None:
             ip_arg_str = f"{addr}/{masklen}"
+            params_dict['ip_arg_str'] = ip_arg_str
         else:
-            raise ipaddress.AddressValueError()
-
-        params_dict['ip_arg_str'] = ip_arg_str
+            raise AddressValueError("IPv6Obj(arg='%s')" % (arg))
 
         return params_dict
 
@@ -1973,7 +2098,7 @@ class IPv6Obj:
         return self.network_object.sixtofour
 
 
-class L4Object:
+class L4Object(object):
     """Object for Transport-layer protocols; the object ensures that logical operators (such as le, gt, eq, and ne) are parsed correctly, as well as mapping service names to port numbers
 
     Examples
@@ -2058,7 +2183,7 @@ class L4Object:
         return f"<L4Object {self.protocol} ports: {crobj.compressed_str}>"
 
 
-class DNSResponse:
+class DNSResponse(object):
     """A universal DNS Response object
 
     Parameters
@@ -2117,6 +2242,7 @@ class DNSResponse:
             )
 
 
+@logger.catch(reraise=True)
 def dns_query(input_str="", query_type="", server="", timeout=2.0):
     """A unified IPv4 & IPv6 DNS lookup interface; this is essentially just a wrapper around dnspython's API.  When you query a PTR record, you can use an IPv4 or IPv6 address (which will automatically be converted into an in-addr.arpa name.  This wrapper only supports a subset of DNS records: 'A', 'AAAA', 'CNAME', 'MX', 'NS', 'PTR', and 'TXT'
 
@@ -2309,6 +2435,7 @@ def dns_query(input_str="", query_type="", server="", timeout=2.0):
     return retval
 
 
+@logger.catch(reraise=True)
 def dns_lookup(input_str, timeout=3, server="", record_type="A"):
     """Perform a simple DNS lookup, return results in a dictionary"""
     rr = Resolver()
@@ -2348,6 +2475,7 @@ def dns_lookup(input_str, timeout=3, server="", record_type="A"):
         }
 
 
+@logger.catch(reraise=True)
 def dns6_lookup(input_str, timeout=3, server=""):
     """Perform a simple DNS lookup, return results in a dictionary"""
     rr = Resolver()
@@ -2372,36 +2500,78 @@ def dns6_lookup(input_str, timeout=3, server=""):
 
 _REVERSE_DNS_REGEX = re.compile(r"^\s*\d+\.\d+\.\d+\.\d+\s*$")
 
+@logger.catch(reraise=True)
+def check_valid_ipaddress(input_addr=None):
+    """
+    Accept an input string with an IPv4 or IPv6 address. If the address is
+    valid, return a tuple of:
+    (input_addr, ipaddr_family)
 
-def reverse_dns_lookup(input_str, timeout=3, server=""):
-    """Perform a simple reverse DNS lookup, return results in a dictionary"""
-    assert _REVERSE_DNS_REGEX.search(input_str), "Invalid address format: '{}'".format(
-        input_str
-    )
-    rr = Resolver()
-    rr.timeout = float(timeout)
-    rr.lifetime = float(timeout)
-    if server:
-        rr.nameservers = [server]
+    Throw an error if the address is not valid.
+    """
+    assert isinstance(input_addr, str)
+    input_addr = input_addr.strip()
+    ipaddr_family = 0
     try:
-        tmp = input_str.strip().split(".")
-        tmp.reverse()
-        inaddr = ".".join(tmp) + ".in-addr.arpa"
-        records = rr.query(inaddr, "PTR")
-        return {
-            "name": records[0].to_text(),
-            "lookup": inaddr,
+        assert IPv4Obj(input_addr)
+        ipaddr_family = 4
+    except Exception as _:
+        pass
+
+    if ipaddr_family == 0:
+        try:
+            assert IPv6Obj(input_addr)
+            ipaddr_family = 6
+        except Exception as _:
+            pass
+
+    error = "FATAL: '{0}' is not a valid IPv4 or IPv6 address.".format(input_addr)
+    assert (ipaddr_family == 4 or ipaddr_family == 6), error
+    return (input_addr, ipaddr_family)
+
+@logger.catch(reraise=True)
+def reverse_dns_lookup(input_str, timeout=3.0, server="", proto="udp"):
+    """Perform a simple reverse DNS lookup on an IPv4 or IPv6 address; return results in a python dictionary"""
+    assert isinstance(proto, str) and (proto=="udp" or proto=="tcp")
+    assert isinstance(float(timeout), float) and float(timeout) > 0.0
+
+    input_str, ipaddr_family = check_valid_ipaddress(input_str)
+
+    tcp_flag = None
+    if proto=="tcp":
+        tcp_flag = True
+    elif proto=="udp":
+        tcp_flag = False
+    else:
+        raise ValueError()
+
+    rr = Resolver()
+    #rr = rr.resolve_address(ipaddr=input_str, tcp=tcp_flag, rdtype="PTR", lifetime=float(timeout))
+    #rr = rr.resolve_address(ipaddr=input_str, tcp=tcp_flag, lifetime=float(timeout))
+
+    if server != "":
+        rr.nameservers = [server]
+
+    records = []
+    retval = {}
+    try:
+        records = rr.resolve_address(ipaddr=input_str, tcp=tcp_flag, lifetime=float(timeout))
+        retval = {
+            "name": records[0].to_text(), # this key was called "addrs"
+            "lookup": input_str,
             "error": "",
-            "addr": input_str,
+            "addrs": [input_str],
         }
     except DNSException as e:
-        return {
-            "addrs": [],
-            "lookup": inaddr,
+        retval = {
+            "name": "",   # this key was called "addrs"
+            "lookup": input_str,
             "error": repr(e),
-            "name": input_str,
+            "addrs": [input_str],
         }
 
+    return retval
+
 
 class CiscoRange(MutableSequence):
     """Explode Cisco ranges into a list of explicit items... examples below...
@@ -2582,7 +2752,8 @@ class CiscoRange(MutableSequence):
     ## Github issue #125
     @property
     def compressed_str(self):
-        """Return a text string with a compressed csv of values
+        """
+        Return a text string with a compressed csv of values
 
         >>> from ciscoconfparse.ccp_util import CiscoRange
         >>> range_obj = CiscoRange('1,3,5,6,7')
